# OrderedQueueWrapper for send and receive messages in order on MNS queue.
# It support only one send process and one receive process.  Message sequence is meaningless if there are multiple send processes and receive processes.
# Note: This code is just a demo solution for how to send and receive messages in sequence on MNS. It's not tested enough. Please do full test if it is applied to production envrionment.

import sys
import os
import time
import threading
from mns.account import Account
from mns.queue import *
import ConfigParser

#Interface of peissit store the seqid.
class PersistStorage:
    def __init__(self, storageType, storageConfig):
        self.storageType = storageType
        self.storageConfig = storageConfig #expected key-value pairs in dictionary.

    def SaveData(self, data):
        pass

    def LoadData(self, data):
        return None

    def GetStorageConfig(self):
        return None

class LocalDiskStorage(PersistStorage):
    def __init__(self, storageConfig):
        PersistStorage.__init__(self, "LocalDisk", storageConfig)
        self.localFileName = ""
        if storageConfig.has_key("localFileName"):
            self.localFileName = storageConfig["localFileName"]
        else:
            self.localFileName = "/tmp/mns_message_seq_id_" + time.strftime("%y%m%d%H%M%S", time.localtime())
            self.storageConfig["localFileName"] = self.localFileName
        
    def SaveData(self, data):
        fh = open(self.localFileName, "w")
        fh.write(str(data))
        fh.close()

    def LoadData(self):
        if not os.path.exists(self.localFileName):
            print "data file %s not exist"  % self.localFileName
            return None
        fh = open(self.localFileName, "r")
        dataStr = fh.read()
        dataStr = dataStr.strip()
        fh.close()
        return dataStr

    def GetStorageConfig(self):
        return self.storageConfig 

class SeqId:
    MAX_ID = sys.maxint
    MIN_ID = 0 
    def __init__(self, init_value, persistStorage = None):
        self.inner_value = init_value
        self.persistStorage = persistStorage
        self.LoadSeqIdFromPersistStorage()
            
    def SaveSeqIdToPersistStorage(self):
        if (self.persistStorage != None):
            self.persistStorage.SaveData(str(self.inner_value))

    def LoadSeqIdFromPersistStorage(self):
        if (self.persistStorage != None):
            data = self.persistStorage.LoadData()
            if (data != None):
                self.inner_value = int(data)

    def SetValue(self, new_value):
        self.inner_value = new_value
        self.SaveSeqIdToPersistStorage()

    def GetValue(self):
        return self.inner_value

    def IncValue(self):
        self.inner_value = self.inner_value + 1
        if (self.inner_value > self.MAX_ID):
            self.inner_value = self.MIN_ID
        self.SaveSeqIdToPersistStorage() 
        return self.inner_value
 
class SeqMessage:
    def __init__(self, seq_id, message):
        self.seq_id = seq_id
        self.message = message

class MessageBuffer:
    def __init__(self, buffer_size, queue_ref, receiveSeqId):
       self.buffer_size = buffer_size
       self.next_seq_id = receiveSeqId

       self.inner_buffer_list = []
       self.queue_ref = queue_ref
       self.MESSAGE_INVISIBALE_TIMEOUT = 60 * 1000 #(60s)
       self.MIN_TIMEOUT = 15 # inseconds
       self.CHECK_INTERVAL_TIME = 15
       self.mutex = threading.Lock() 

    def InsertMessage(self, seq_message):
       if(self.buffer_size <= len(self.inner_buffer_list)):
           print "InserMessage ERROR: message buffer is full"
           return -1
      
       if seq_message.seq_id.GetValue() < self.next_seq_id.GetValue():  
           #there is a bug when seq id is near the sys.maxint. but if sys.maxint is 64bit int which is big enough to avoid this bug.
           print "Receive message which has been received before."
           return 1
       if self.mutex.acquire():
           pos = 0
           for smsg in self.inner_buffer_list:
               if (seq_message.seq_id.GetValue() > smsg.seq_id.GetValue()):
                   pos = pos + 1
               else:
                   break
           self.inner_buffer_list.insert(pos, seq_message) #need lock before insert.
           self.mutex.release()

       return 0

    def GetMessage(self):
       if (len(self.inner_buffer_list) <=0):
           return None

       headMsg = self.inner_buffer_list[0]
       #print "debug: headMsg seq id is %d, expected next id is %d" % (headMsg.seq_id.GetValue(), self.next_seq_id.GetValue())
       if (headMsg.seq_id.GetValue() == self.next_seq_id.GetValue()):
           self.next_seq_id.IncValue()
           if self.mutex.acquire():
               del self.inner_buffer_list[0]
               self.mutex.release()
           return headMsg
       return None
       
    def GetNextSeqId(self):
       return self.next_seq_id
    
    def CleanBuffer(self):
       if self.mutex.acquire():
           self.inner_buffer_list = []
           self.next_seq_id = SeqId(SeqId.MIN_ID)
           self.mutex.release()
        
    def KeepMessageInVisiable(self):
       if self.mutex.acquire():
           for smsg in self.inner_buffer_list:
               innerMsg = smsg.message
               try:
                   if(((innerMsg.next_visible_time/1000) - time.time()) <= self.MIN_TIMEOUT):
                       print "need change message visiblity time"
                       self.queue_ref.change_message_visibility(innerMsg.receipt_handle, self.MESSAGE_INVISIBALE_TIMEOUT)
               except MNSExceptionBase, e:
                   print "Change Message Visibility Fail!\nException:%s\n\n" % e
           self.mutex.release()

class CheckTimeoutThread(threading.Thread):
    def __init__(self, messageBuffer):
        threading.Thread.__init__(self)
        self.msgBuffer = messageBuffer
        self.isRunning = True

    def run(self):
        while(self.isRunning):
            self.msgBuffer.KeepMessageInVisiable()
            time.sleep(self.msgBuffer.CHECK_INTERVAL_TIME)
          
class OrderedQueueWrapper:
    def __init__(self, queue, msgBufferSize = 10000, sendInitSeqId = 0, receiveInitSeqId = 0, sendSeqIdPersistStroage = None, receiveSeqIdPersistStorage = None): 
        self.raw_queue = queue
        self.msgBuffer = MessageBuffer(msgBufferSize, receiveInitSeqId, self.raw_queue)
        self.sendSeqId = SeqId(sendInitSeqId, sendSeqIdPersistStroage)
        self.receiveSeqId = SeqId(receiveInitSeqId, receiveSeqIdPersistStorage)
        self.msgBuffer = MessageBuffer(msgBufferSize, self.raw_queue, self.receiveSeqId)
        self.checkThread = CheckTimeoutThread(self.msgBuffer)
        self.checkThread.setDaemon(True)
        self.checkThread.start()
    
    def GetRawQueue(self):
        return self.raw_queue

    def GetMsgBuffer(self):
        return self.msgBuffer

    def AddSeqIdToMessageBody(self, message, seqId):
        idStr = "#%s#" % str(seqId.GetValue())
        message.message_body = idStr + message.message_body

    def ParseSeqIdFromMessageBody(self, message):
        if len(message.message_body) <= 0:
            print "Error message body is empty"
            return None
        if (message.message_body[0] != '#'):
            print "Error message is not start with valid seq id"
            return None
        pos = message.message_body.find("#", 1, len(message.message_body))
        if (pos > 1):
            idStr = message.message_body[1:pos]
            #print "debug: seqid is %s" % str(idStr)
            seqId = int(idStr) # need to handle exception while convert str to int.
            message.message_body = message.message_body[(pos + 1):]
            #print "debug: message body is %s" % message.message_body
            return SeqId(seqId)             
        else:
            return None

    def SendMessageInOrder(self, message):
        self.AddSeqIdToMessageBody(message, self.sendSeqId)
        self.raw_queue.send_message(message)
        self.sendSeqId.IncValue()

    def ReceiveMessageInOrder(self, wait_seconds):
        headMsg = self.msgBuffer.GetMessage()
        while (headMsg == None):
            recMsg = self.raw_queue.receive_message(wait_seconds)
            seqId = self.ParseSeqIdFromMessageBody(recMsg)
            seqMessage = SeqMessage(seqId, recMsg)
            res = self.msgBuffer.InsertMessage(seqMessage)
            if (res == 0):
                headMsg = self.msgBuffer.GetMessage()
            elif(res == 1):# this is a message that has been returned before due to failover, return to callder directly.
                headMsg = seqMessage
            elif(res == -1):
                return -1 # message buffer is full, let caller decide what to do. one action is clean the buffer, or enlarge the buffer size.
        return headMsg.message

    def StopCheckThread(self):
        self.checkThread.isRunning = False

### Test Code ###
def testOrderedQueueWrapper():
    #create queue
    endpoint = ""
    accessKeyId = ""
    accessKeySecret = ""

    if (endpoint == "" or accessKeyId == "" or accessKeySecret == ""):
        print "please config the endpoint, accessKeyId, accessKeySecret before run test"
        sys.exit(1)

    myAccount = Account(endpoint, accessKeyId, accessKeySecret)
    myQueue = myAccount.get_queue("MNS-OrderedQueueWrapperTest-%s" % time.strftime("%y%m%d-%H%M%S", time.localtime()))
    queueMeta = QueueMeta()
    queueMeta.set_visibilitytimeout(100)
    queueMeta.set_maximum_message_size(10240)
    queueMeta.set_message_retention_period(3600)
    queueMeta.set_delay_seconds(10)
    queueMeta.set_polling_wait_seconds(20)
    try:
        queueUrl = myQueue.create(queueMeta)
        sys.stdout.write("Create Queue Succeed!\nQueueURL:%s\n\n" % queueUrl)
    except MNSExceptionBase, e:
        sys.stderr.write("Create Queue Fail!\nException:%s\n\n" % e)
        sys.exit(1)

    print "Test with common queue:"
    #send 10 messages with common queue
    for i in range(1, 11):
        msg_body = str(i)
        message = Message(msg_body)
        message.set_delayseconds(0)
        message.set_priority(10)
        try:
            send_msg = myQueue.send_message(message)
            print "Send Message Succeed with commont queue. message body is :%s" % msg_body
        except MNSExceptionBase, e:
            sys.stderr.write("Send Message Fail!\nException:%s\n\n" % e)
            sys.exit(1)

    #receive 10 messages with common queue
    for i in range(1,11):
        try:
            wait_seconds = 10
            recv_msg = myQueue.receive_message(wait_seconds)
            print "Receive Message Succeed with common queue! message body is %s" % recv_msg.message_body
            myQueue.delete_message(recv_msg.receipt_handle)
        except MNSExceptionBase, e:
            sys.stderr.write("Receive Message Fail!\nException:%s\n\n" % e)
            sys.exit(1)

    print "\nTest with OrderedQueueWrapper:"

    #create oredered queue wrapper
    sendSeqIdConfig = {"localFileName":"/tmp/mns_send_message_seq_id"}
    sendSeqIdPS = LocalDiskStorage(sendSeqIdConfig)
    receiveSeqIdConfig = {"localFileName":"/tmp/mns_receive_message_seq_id"}
    receiveSeqIdPS = LocalDiskStorage(receiveSeqIdConfig)
   
    orderedQueue = OrderedQueueWrapper(myQueue, sendSeqIdPersistStroage = sendSeqIdPS, receiveSeqIdPersistStorage = receiveSeqIdPS)

    #send 10 messages with ordered queue
    for i in range(1, 11):
        msg_body = str(i)
        message = Message(msg_body)
        message.set_delayseconds(0)
        message.set_priority(10)
        print "orderedqueuewrapper: send message body %s" % message.message_body
        orderedQueue.SendMessageInOrder(message)

    #receive 10 messages with orderd queue
    for i in range(1, 11):
        wait_seconds = 10
        recv_msg = orderedQueue.ReceiveMessageInOrder(wait_seconds)
        print "orderedqueuewrapper: receive message body is %s" % recv_msg.message_body
        myQueue.delete_message(recv_msg.receipt_handle)
    #delete queue
    orderedQueue.StopCheckThread()  #IMPORTANT!!! close the checkThread of orderedqueue.
    myQueue.delete()

if __name__ == "__main__":
    testOrderedQueueWrapper() 
